# encoding: utf-8

import json

class Node(object):


    def __init__(self, parent=None, key=None):

        self.reset()

        self.parent   = parent
        self.key      = key



    def reset(self):

        self.setas_root()

        self.key      = None
        self.value    = 0

        self.setas_leaf()



    def setas_root(self):

        self.parent   = None



    def setas_leaf(self):

        self.children = {}



    def root(self):

        return self.parent is None



    def leaf(self):

        return not bool(self.children)



    def meaningless(self):

        return 0 == self.value



    # def __str__(self):
    #     return str(self.key)



    def __repr__(self):

        return ', '.join((str(id(self.parent)),
                         str(self.key),
                         str(self.value),
                         str(self.root()),
                         str(self.leaf()),
                         json.dumps({
                             k:id(v) for k, v in self.children.items()
                         }, ensure_ascii=False)))

        # return json.dumps(self.__dict__, ensure_ascii=False, indent=4)



# encoding: utf-8

import progressbar
import _pickle as pickle

from trie import TrieHeap


class TrieDouble(object):


    def __init__(self, ratio=None, batch_size=1):

        self.ratio      = ratio

        self._trie_top  = TrieHeap()
        self._trie_rest = TrieHeap()

        self._count     = 0
        self.batch_size = batch_size



    def check(self):

        # TODO:

        pass



    @property
    def need_tune(self):

        return 0 == self._count % self.batch_size



    def trie_to_trie(self, node, trie1, trie2):
        """ Delete node of trie1, Add to trie2

        Parameters
        ----------
        node: trie.heap.node in trie1

        trie1: trie

        trie2: trie
        """

        # update in trie2
        infoes = set()

        for n in node.infoes:
            key       = trie1.trace(n)
            value     = node.key
            node_trie = trie2.add(key, value)
        
            infoes.add(node_trie)
        
        # delete in trie1 & heap1
        trie1.delete_bynode(node)
        
        # update in heap2
        node.infoes = infoes
        trie2._heap.add_bynode(node)



    def tune_top_ratio(self):

        if (self._trie_top.count_sum-self._trie_top.count_min)/\
           (self._trie_top.count_sum+self._trie_rest.count_sum) >= self.ratio:

            self.trie_to_trie(self._trie_top.node_min,
                              self._trie_top,
                              self._trie_rest)



    def tune_rest_ratio(self):

        if self._trie_top.count_sum/\
          (self._trie_top.count_sum+self._trie_rest.count_sum) < self.ratio:

            self.trie_to_trie(self._trie_rest.node_max,
                              self._trie_rest,
                              self._trie_top)


    def tune_top_unique(self, node):

        if self.have_rest_heap(node.key):
            self.trie_to_trie(node, self._trie_top, self._trie_rest)



    def tune_rest_unique(self, node):
        """ for key unique

        warnging: key is change with each update.
        """

        if self.have_top_heap(node.key):
            self.trie_to_trie(node, self._trie_rest, self._trie_top)



    def update(self, key, value=1):

        self._count += 1

        if self._trie_top.empty:
            self._trie_top.update(key, value)

        elif self.have_top_trie(key):
            _, node_heap = self._trie_top.update(key, value)

            if self.need_tune:
                self.tune_top_unique(node_heap)
                self.tune_top_ratio()

        elif self.have_rest_trie(key):
            _, node_heap = self._trie_rest.update(key, value)

            if self.need_tune:
                self.tune_rest_unique(node_heap)
                self.tune_rest_ratio()

        elif self._trie_rest.empty:
            # rest empty --> top <= ratio
            # put more, then tune
            _, node_heap = self._trie_top.update(key, value)

            if self.need_tune:
                self.tune_top_unique(node_heap)
                self.tune_top_ratio()

        else:
            _, node_heap = self._trie_rest.update(key, value)

            if self.need_tune:
                self.tune_rest_unique(node_heap)
                self.tune_rest_ratio()



    def have_top_trie(self, key):

        return self._trie_top.have_trie(key)



    def have_rest_trie(self, key):

        return self._trie_rest.have_trie(key)



    def have_top_heap(self, key):

        return self._trie_top.have_heap(key)



    def have_rest_heap(self, key):

        return self._trie_rest.have_heap(key)



    def save(self, fn, predict=False):

        if not predict:
            with open(fn, 'wb') as f:
                pickle.dump(self.__dict__, f)

        else:
            with open('%s.top' % fn, 'wb') as f:
                pickle.dump({
                                'ratio'    : self.ratio,
                                '_trie_top': self._trie_top
                            }, f)



    def load(self, fn, predict=False):

        if not predict:
            with open(fn, 'rb') as f:
                d = pickle.load(f)
                self.__dict__.update(d)

        else:
            with open('%s.top' % fn, 'rb') as f:
                d = pickle.load(f)
                self.__dict__.update(d)



# encoding: utf-8

import sys
import traceback
import progressbar
import _pickle as pickle

from node              import Node
from collections       import defaultdict
from heap.heap_brother import HeapBrotherPairMax


class Trie(object):


    def __init__(self):

        self._tree    = Node()
        self._counter = defaultdict(int)



    def add(self, key, value=1):

        node_current = self._tree
        length       = len(key) - 1

        for i, w in enumerate(key):
            # child node
            node_child = node_current.children.get(w, Node(parent=node_current, key=w))

            if length == i:
                node_child.value += value

            # current node
            node_current.children[w] = node_child

            # update current ref
            node_current = node_child

        return node_child




    def __getitem__(self, key):

        node_current = self._tree

        for w in enumerate(key):

            # check
            node = node_current.children.get(w, None)

            if node is None:
                return node

            # update current ref
            node_current = node

        return node



    def delete(self, key):

        node = self[key]
        self.delete_bynode(node)




    def delete_bynode(self, node):

        if node:

            if node.root():
                node.reset()

            elif node.leaf():

                # record node.parent
                node_parent = node.parent
                
                # delete node
                # revise node.parent info
                node_parent.children.pop(node.key)

                # mark node.parent as node
                node = node_parent

                while(node.leaf() and node.meaningless()):
                    # record node.parent
                    node_parent = node.parent

                    # delete node
                    # revise node.parent info
                    node_parent.children.pop(node.key)

                    # border condition, root
                    if node_parent.root():
                        break

                    # mark node.parent as node
                    node = node_parent

            else:
                node.value = 0



    def have(self, key):

        node_current = self._tree

        for w in key:
            if w not in node_current.children.keys():
                return False
            node_current = node_current.children[w]

        if 0 == node_current.value:
            return False

        return True



    def trace(self, node):

        t = []

        n = node
        while(n.parent is not None):
            t.append(n.key)
            n = n.parent

        return ''.join(t[::-1])



    def traverse_broad(self):

        current_nodes = [self._tree]

        cnt = 0
        bar = progressbar.ProgressBar(max_value=progressbar.UnknownLength)
        while(current_nodes):
            # visit current node
            node = current_nodes.pop(0)
            yield(self.trace(node), node.value)

            # add children
            current_nodes.extend(list(node.children.values()))

            cnt += 1
            bar.update(cnt)



    def revise(self, key, cnt):

        # TODO:

        pass



    @property
    def counter(self):

        self._counter = defaultdict(int)

        for k, v in self.traverse_broad():
            self._counter[v] += 1

        try:
            self._counter.pop(0)
        except:
            traceback.print_exc()

        return self._counter



    def save(self, fn):

        with open(fn, 'wb') as f:
            pickle.dump(self.__dict__, f)


    def load(self, fn):

        with open(fn, 'rb') as f:
            d = pickle.load(f)
            self.__dict__.update(d)




class TrieHeap(Trie):


    def __init__(self):

        super(TrieHeap, self).__init__()
        self._heap = HeapBrotherPairMax()



    def have_trie(self, key):

        return self.have(key)




    def have_heap(self, key):

        return self._heap.have(key)



    def update(self, key, value=1):

        # update trie
        node_trie = self.add(key, value)

        # update heap
        if 1 == node_trie.value:
            key       = node_trie.value
            value     = 1
            info      = node_trie
            node_heap = self._heap.update_byinfo(key, value, info)
        else:
            # delete info old
            key   = node_trie.value - 1
            value = 1
            info  = node_trie
            node_delete = self._heap.delete_byinfo(key, value, info)

            # delete node when empty
            if node_delete.empty:
                self._heap.delete_bynode(node_delete)

            # add new
            key       = node_trie.value
            value     = 1
            info      = node_trie
            node_heap = self._heap.update_byinfo(key, value, info)

        return node_trie, node_heap




    def delete_bynode(self, node):
        """
        Parameters
        ----------
        node : node of heap
        """

        # delete trie
        for n in node.infoes:
            super(TrieHeap, self).delete_bynode(n)

        # delete heap
        self._heap.delete_bynode(node)



    @property
    def count_max(self):

        if self.empty:
            return 0

        node = self.node_max

        return node.key * node.value



    @property
    def count_min(self):

        if self.empty:
            return 0

        node = self.node_min

        return node.key * node.value



    @property
    def count_sum(self):

        return self._heap.value_sum



    @property
    def empty(self):

        return self._heap.empty


    @property
    def node_max(self):

        return self._heap.node_max


    @property
    def node_min(self):

        return self._heap.node_min
